from pathlib import Path
import os
import yaml
from configparser import ConfigParser
from typing import Optional
from ansible_workspace.models import AnsibleRole
from ansible_workspace.git import GitRepo
from ansible_workspace.output import console


def find_ansible_roles_path(ansible_cfg: Path) -> Path:
    """Find the path to the ansible roles directory."""
    if os.getenv("ANSIBLE_ROLES_PATH"):
        return Path(str(os.getenv("ANSIBLE_ROLES_PATH")))

    if Path.cwd().joinpath("roles").exists():
        return Path.cwd().joinpath("roles")

    if ansible_cfg.exists():
        config = parse_ansible_cfg(ansible_cfg)
        if config.has_option("defaults", "roles_path"):
            return Path(config["defaults"]["roles_path"])

    return Path().home().joinpath(".ansible", "roles")


def get_roles_config(source: Optional[Path] = None) -> list[AnsibleRole]:
    """Find the roles config file."""
    if source:
        config_path = Path(source)
    else:
        for path in Path.cwd().rglob("requirements.y*ml"):
            if not 'collections' in str(path):
                config_path = Path(path)
                break
        else:
            raise ValueError("Could not find the roles config file")

    config = yaml.safe_load(config_path.read_text())
    if isinstance(config, dict) and config.get("roles"):
        config = config["roles"]

    roles = []
    for role in config:
        if (
            isinstance(role, dict)
            and role.get("name")
            and role.get("src")
            and role.get("scm") == "git"
        ):
            roles.append(AnsibleRole(**role))
    return roles


def handle_roles(
    roles: list[AnsibleRole], roles_path: Path, symlink_path: Optional[Path] = None
) -> None:
    for role in roles:
        repo = GitRepo(role.src, roles_path.joinpath(role.name))
        if not roles_path.exists():
            roles_path.mkdir(parents=True)
        if not roles_path.joinpath(role.name).joinpath('.git').exists():
            repo.clone()
        repo.fetch_all()

        if role.version:
            if role.version in repo.get_tags():
                repo.checkout(role.version)
            elif role.version in repo.get_local_branches():
                repo.checkout(f'checkout {role.version}')
            elif f'origin/{role.version}' in repo.get_remote_branches():
                repo.checkout_remote_branch(role.version)
        else:
            if 'master' in repo.get_local_branches():
                repo.checkout('master')
            elif 'main' in repo.get_local_branches():
                repo.checkout('main')

        if symlink_path:
            symlink_role = symlink_path.joinpath(role.name)
            if symlink_role.exists():
                symlink_role.unlink()
            console.print(f"Symlinking \"{role.name}\" to \"{symlink_role}\"")
            symlink_role.symlink_to(repo.path)


def parse_ansible_cfg(path: Path) -> ConfigParser:
    """Parse the ansible.cfg file."""
    if not path.exists():
        raise ValueError("Could not find the ansible.cfg file")

    config = ConfigParser()
    config.read(str(path))
    return config


def add_roles_to_gitignore(roles: list[AnsibleRole], roles_path: Path) -> None:
    """Add the roles to the .gitignore file."""
    gitignore = Path.cwd().joinpath(".gitignore")
    if not gitignore.exists():
        gitignore.touch()
    for role in roles:
        cwd = Path.cwd()
        role_path = roles_path.joinpath(role.name)
        if not role.name in gitignore.read_text():
            with open(gitignore, "a") as f:
                relative_path = role_path.relative_to(cwd)
                console.print(f"Adding \"{relative_path}\" to .gitignore")
                f.write(f"{relative_path}/\n")



from subprocess import check_output, STDOUT, CalledProcessError, Popen, PIPE
from ansible_workspace.output import console
from pathlib import Path


class GitRepo:
    def __init__(self, source: str, path: Path):
        self.source = source
        self.path: Path = path

    def run(self, *args) -> str:
        try:
            process = Popen(
                ('git', '-C', str(self.path), *args), stdout=PIPE, stderr=PIPE
            )
            stdout, stderr = process.communicate()
            if process.returncode != 0:
                raise CalledProcessError(
                    process.returncode, process.args, output=stdout, stderr=stderr
                )
            return stdout.decode('utf-8')
        except CalledProcessError as e:
            console.print(e.output.decode('utf-8'))
            console.print(e.stderr.decode('utf-8'))
            raise ValueError(f"Failed to run git command: {e.cmd}") from e

    def checkout(self, branch: str) -> str:
        console.print(f"Checking out \"{branch}\" on \"{self.path}\"")
        return self.run('checkout', branch)

    def checkout_remote_branch(self, branch: str) -> str:
        console.print(
            f"Checking out \"{branch}\" from \"origin/{branch}\" on \"{self.source}\""
        )
        return self.run('checkout', '-b', branch, f'origin/{branch}')

    def clone(self) -> str:
        if not self.path.exists():
            self.path.mkdir(parents=True)
        console.print(f"Cloning \"{self.source}\" to \"{self.path}\"")
        return self.run('clone', self.source, '.')

    def fetch_all(self) -> str:
        return self.run('fetch', '--all')

    def get_remote_branches(self) -> list[str]:
        output = self.run('branch', '-r').splitlines()
        return [r.replace('*', '').strip() for r in output]

    def get_local_branches(self) -> list[str]:
        output = self.run('branch').splitlines()
        return [r.replace('*', '').strip() for r in output]

    def get_tags(self) -> list[str]:
        return self.run('tag').splitlines()



from dataclasses import dataclass, field
from copy import copy
from typing import Any


def default_field(obj):
    return field(default_factory=lambda: copy(obj))


@dataclass
class VSCodeFolder:
    name: str
    path: str


@dataclass
class VSCodeWorkspace:
    folders: list[VSCodeFolder]
    settings: dict[str, str] = default_field({})
    extensions: dict[str, Any] = default_field(
        {
            "recommendations": [
                "redhat.ansible",
            ]
        }
    )



from dataclasses import dataclass


@dataclass
class TmuxpWindow:
    window_name: str
    start_directory: str
    layout: str = 'tiled'
    shell_command_before: str = ''


@dataclass
class TmuxpWorkspace:
    session_name: str
    start_directory: str
    windows: list[TmuxpWindow]



import typer
from typing import Optional
from pathlib import Path
from ansible_workspace.models import AnsibleRole
from ansible_workspace.output import console
from ansible_workspace.misc import (
    find_ansible_roles_path,
    get_roles_config,
    handle_roles,
    add_roles_to_gitignore,
)
from ansible_workspace.tmuxp import TmuxpWorkspace, TmuxpWindow
from ansible_workspace.vscode import (
    VSCodeFolder,
    VSCodeWorkspace,
)
import importlib.metadata
import json
import yaml
from dataclasses import asdict

app = typer.Typer()


def general(
    roles_config: Optional[Path],
    ansible_config: Path,
    ansible_roles_path: Path,
    symlink_path: Optional[Path],
) -> list[AnsibleRole]:
    roles_path = ansible_roles_path or find_ansible_roles_path(ansible_config)

    roles = get_roles_config(roles_config)
    handle_roles(roles, roles_path, symlink_path)

    if str(Path().cwd()) in str(roles_path) and Path().cwd().joinpath(".git").exists():
        add_roles_to_gitignore(roles, roles_path)

    if (
        symlink_path
        and str(Path().cwd()) in str(symlink_path)
        and Path().cwd().joinpath(".git").exists()
    ):
        add_roles_to_gitignore(roles, symlink_path)
    return roles


@app.command()
def tmuxp(
    roles_config: Optional[Path] = typer.Option(
        None, "--roles-config", "-r", help="Path to roles config file"
    ),
    ansible_config: Path = typer.Option(
        Path.cwd().joinpath("ansible.cfg"), help="Path to ansible.cfg file"
    ),
    roles_path: Optional[Path] = typer.Option(
        None, envvar="ANSIBLE_ROLES_PATH", help="Path to the roles directory"
    ),
    symlink_path: Optional[Path] = typer.Option(
        None, help="Path to symlink the roles to"
    ),
    workspaces_path: Path = typer.Option(
        Path.home().joinpath("workspaces"),
        help="Path to the workspace directory",
        envvar='WORKSPACES_PATH',
    ),
):
    ansible_roles_path = roles_path or find_ansible_roles_path(ansible_config)
    roles = general(roles_config, ansible_config, ansible_roles_path, symlink_path)
    windows = []
    for role in roles:
        windows.append(
            TmuxpWindow(
                window_name=role.name,
                start_directory=str(ansible_roles_path.joinpath(role.name)),
            )
        )

    project_name = Path.cwd().name
    workspace = TmuxpWorkspace(
        session_name=project_name,
        start_directory=str(Path.cwd()),
        windows=[
            TmuxpWindow(
                window_name=project_name,
                start_directory=str(Path.cwd()),
            ),
            TmuxpWindow(
                window_name="ansible",
                start_directory=str(roles_path),
            ),
        ]
        + windows,
    )
    if not workspaces_path.exists():
        workspaces_path.mkdir(parents=True)
    config_path = workspaces_path.joinpath(f"{project_name}.tmuxp-workspace.yml")
    with open(config_path, "w") as f:
        yaml.safe_dump(asdict(workspace), f)
    console.print(f"Created workspace config at \"{config_path}\"")


@app.command()
def vscode(
    roles_config: Optional[Path] = typer.Option(
        None, "--roles-config", "-r", help="Path to roles config file"
    ),
    ansible_config: Path = typer.Option(
        Path.cwd().joinpath("ansible.cfg"), help="Path to ansible.cfg file"
    ),
    roles_path: Optional[Path] = typer.Option(
        None, envvar="ANSIBLE_ROLES_PATH", help="Path to the roles directory"
    ),
    symlink_path: Optional[Path] = typer.Option(
        None, help="Path to symlink the roles to"
    ),
    workspaces_path: Path = typer.Option(
        Path.home().joinpath("workspaces"),
        help="Path to the workspace directory",
        envvar='WORKSPACES_PATH',
    ),
):
    ansible_roles_path = roles_path or find_ansible_roles_path(ansible_config)
    roles = general(roles_config, ansible_config, ansible_roles_path, symlink_path)
    windows = []
    for role in roles:
        windows.append(
            VSCodeFolder(
                name=role.name,
                path=str(ansible_roles_path.joinpath(role.name)),
            )
        )

    project_name = Path.cwd().name
    workspace = VSCodeWorkspace(
        folders=[
            VSCodeFolder(
                name=project_name,
                path=str(Path.cwd()),
            ),
        ]
        + windows,
    )
    if not workspaces_path.exists():
        workspaces_path.mkdir(parents=True)
    config_path = workspaces_path.joinpath(f"{project_name}.code-workspace")
    with open(config_path, "w") as f:
        json.dump(asdict(workspace), f)
    console.print(f"Created workspace config at \"{config_path}\"")


@app.command()
def version():
    console.print(importlib.metadata.version('ansible-workspace'))



from setuptools import setup
from os import path
from subprocess import check_output

with open("./requirements.txt") as f:
    required = f.read().splitlines()

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

version = '1.0.0'

try:
    version = (
        check_output(['git', 'describe', '--tags']).strip().decode().replace('v', '')
    )
except:
    pass

setup(
    name='ansible_workspace',
    version=version,
    description='Create a workspace for multiple tools to easier develop ansible playbooks with roles.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='rwxd',
    author_email='rwxd@pm.me',
    url="https://github.com/rwxd/ansible_workspace",
    license='MIT',
    packages=['ansible_workspace'],
    install_requires=required,
    entry_points={
        "console_scripts": ["ansible-workspace = ansible_workspace.__main__:main"]
    },
    classifiers=[],
)
